import datetime
from typing import List, Dict

import pytest
import coreschema

from apistar import Route
from apistar.test import TestClient
from apistar.handlers import serve_schema
from apistar.interfaces import Schema
from pydantic import BaseModel
from coreapi import Field, Link
from coreapi.codecs import CoreJSONCodec

from apistar_pydantic import (
    ASyncIOApp, WSGIApp, QueryData, FormData, BodyData,
    JSONRenderer
)


class Model(BaseModel):
    integer: int
    text: str

    def compute(self):
        return {
            'integer': 10 * self.integer,
            'text': self.text.upper()
        }


def client_factory(app_class, routes):
    app = app_class(
        settings={
            'RENDERERS': [JSONRenderer()]
        },
        routes=routes
    )
    return TestClient(app)


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_url(app_class):
    def url_argument_view(arg1: str):
        return arg1.upper()

    client = client_factory(app_class, [
        Route('/url_argument/{arg1}', 'GET', url_argument_view),
    ])
    res = client.get('/url_argument/aaa')
    assert res.json() == 'AAA'


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_query_simple(app_class):
    def query_argument_simple_view(arg1: int):
        return arg1 * 10

    client = client_factory(app_class, [
        Route('/query_argument_simple', 'GET', query_argument_simple_view),
    ])
    res = client.get('/query_argument_simple', params={'arg1': 10})
    assert res.json() == 100


args = {
    'integer': 10,
    'text': 'abc'
}
expected = {
    'integer': 100,
    'text': 'ABC'
}


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_query_model(app_class):

    def query_argument_view(model: QueryData[Model]):
        return model.compute()

    client = client_factory(app_class, [
        Route('/query_argument', 'GET', query_argument_view)
    ])

    res = client.get('/query_argument', params=args)
    assert res.json() == expected


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_form_model(app_class):
    class FormModel(BaseModel):
        integer: List[int]
        text: List[str]

        def compute(self):
            return {
                'integer': 10 * self.integer[0],
                'text': self.text[0].upper()
            }

    def form_argument_view(model: FormData[FormModel]):
        return model.compute()

    client = client_factory(app_class, [
        Route('/form_argument', 'POST', form_argument_view)
    ])

    res = client.post('/form_argument', data=args)
    assert res.json() == expected


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_body_model(app_class):

    def body_argument_view(model: BodyData[Model]):
        return model.compute()

    client = client_factory(app_class, [
        Route('/body_argument', 'PUT', body_argument_view)
    ])
    res = client.put('/body_argument', json=args)
    assert res.json() == expected


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_mixed_arguments(app_class):

    def mixed_arguments_view(query: QueryData[Model],
                             body: BodyData[Model]):
        return query.integer * body.integer * query.text

    client = client_factory(app_class, [
        Route('/mixed_arguments', 'POST', mixed_arguments_view)
    ])

    res = client.post(
        '/mixed_arguments',
        params={
            'integer': 2,
            'text': 'a'
        },
        json={
            'integer': 2,
            'text': ''
        })
    assert res.json() == 'aaaa'


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_renderer(app_class):

    def render() -> Model:
        return Model(
            integer=100,
            text='ABC'
        )
    client = client_factory(app_class, [
        Route('/render', 'GET', render)
    ])

    res = client.get('/render')
    assert res.json() == expected


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_renderer_mixed(app_class):

    def render() -> Dict[str, List[Model]]:
        return {
            'model': [Model(integer=100, text='ABC')]
        }

    client = client_factory(app_class, [
        Route('/render', 'GET', render)
    ])

    res = client.get('/render')
    assert res.json() == {
        'model': [expected]
    }


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_renderer_datetime(app_class):

    now = datetime.datetime.now()

    def render() -> datetime.datetime:
        return now

    client = client_factory(app_class, [
        Route('/render', 'GET', render)
    ])

    res = client.get('/render')
    assert res.json() == now.timestamp()


@pytest.mark.parametrize('app_class', [ASyncIOApp, WSGIApp])
def test_schema(app_class):

    def add_model(model: BodyData[Model]):
        """
            add_model description
        """
        return model.integer

    def list_models():
        """list_models description"""
        return

    def show_model(id: int):
        """show_model description"""
        return

    def show_model2(id: int):
        """show_model2 description"""
        return

    app = app_class(
        routes=[
            Route('/add_model', 'POST', add_model),
            Route('/list_models', 'GET', list_models),
            Route('/show_model', 'GET', show_model),
            Route('/show/model/{id}', 'GET', show_model2),
            Route('/schema', 'GET', serve_schema),
        ],
        settings={
            'SCHEMA': {'TITLE': 'My API'}
        }
    )
    client = TestClient(app)

    schema = Schema(title='My API', url='/schema/', content={
        'add_model': Link(
            url='/add_model',
            action='POST',
            description='add_model description',
            fields=[
                Field(
                    name='integer',
                    location='form',
                    required=True,
                    schema=coreschema.Integer()
                ),
                Field(
                    name='text',
                    location='form',
                    required=True,
                    schema=coreschema.String()
                )
            ]
        ),
        'list_models': Link(
            url='/list_models',
            action='GET',
            description='list_models description',
            fields=[]
        ),
        'show_model': Link(
            url='/show_model',
            action='GET',
            description='show_model description',
            fields=[
                Field(
                    name='id',
                    location='query',
                    required=True,
                    schema=coreschema.Integer()
                )
            ]
        ),
        'show_model2': Link(
            url='/show/model/{id}',
            action='GET',
            description='show_model2 description',
            fields=[
                Field(
                    name='id',
                    location='path',
                    required=True,
                    schema=coreschema.Integer()
                )
            ]
        )
    })

    res = client.get('/schema')
    assert res.status_code == 200

    codec = CoreJSONCodec()
    document = codec.decode(res.content)

    assert document.url == '/schema'
    assert document.title == schema.title
    for name, link in schema.links.items():
        assert name in document
        assert link.url == document[name].url
        assert link.action == document[name].action
        assert link.description == document[name].description
        for expected, result in zip(sorted(link.fields), sorted(document[name].fields)):
            assert expected.__class__ == result.__class__




from typing import Dict

from apistar import Route
from apistar_pydantic import (
    WSGIApp as App,
    QueryData, BodyData, FormData
)
from pydantic import BaseModel
from apistar.handlers import (
    serve_schema, serve_static, api_documentation, javascript_schema
)

#
# Declare models
#


class City(BaseModel):
    name: str
    population: int


class Computer(BaseModel):
    model: str
    price: float


#
# Create views
#

def resource_query(city: QueryData[City]) -> str:
    return "%s has %d citizens." % (city.name, city.population)


def resource_body(computer: BodyData[Computer]) -> str:
    return "%s costs R$ %.2f" % (computer.model, computer.price)


def resource_mixed(city: QueryData[City],
                   computer: BodyData[Computer]) -> Dict[str, BaseModel]:
    return {
        'city': city,
        'computer': computer
    }

#
# Start the app
#

app = App(
    routes=[
        Route('/resource_query', 'GET', resource_query),
        Route('/resource_body', 'POST', resource_body),
        Route('/resource_mixed', 'POST', resource_mixed),

        Route('/docs', 'GET', api_documentation),
        Route('/{path}', 'GET', serve_static),
        Route('/schema/', 'GET', serve_schema),
        Route('/schema.js', 'GET', javascript_schema),
    ]
)

app.main()



import enum
import datetime
import inspect
import json
import textwrap
from typing import (
    Any, Callable, Tuple, Optional, Mapping, List, Set, Type, cast
)

import pydantic
import coreapi
import uritemplate
import coreschema
from apistar import exceptions, http, Settings, Route, Component
from apistar.core import flatten_routes
from apistar.interfaces import (
    Injector, Schema, Router, Templates, StaticFiles, CommandLineClient,
    Console, SessionStore
)
from apistar.components import (
    dependency, templates, statics, router, commandline, console,
    sessions
)
from apistar.frameworks.wsgi import WSGIApp
from apistar.frameworks.asyncio import ASyncIOApp
from apistar.renderers import JSONRenderer
from apistar.types import (
    ParamName, KeywordArgs, ParamAnnotation, WSGIEnviron, Handler, UMIChannels,
    UMIMessage, RouteConfig
)

__all__ = [
    'ASyncIOApp', 'WSGIApp',
    'JSONRenderer',
    'QueryData', 'FormData', 'BodyData'
]


class HTTPParameterData:
    _request_location: str = None
    _schema_location: str = None

    def __getitem__(self, type_cls):
        return type(type_cls.__name__, (type_cls, self.__class__), {})


class _QueryData(HTTPParameterData):
    _request_location: str = 'query'
    _schema_location: str = 'query'


class _FormData(HTTPParameterData):
    _data_location: str = 'form'
    _schema_location: str = 'form'


class _BodyData(HTTPParameterData):
    _data_location: str = 'body'
    _schema_location: str = 'form'


QueryData = _QueryData()
FormData = _FormData()
BodyData = _BodyData()


class PydanticHTTPResolver(dependency.HTTPResolver):
    """
    Handles resolving parameters for HTTP requests with pydantic.
    """

    @classmethod
    def coerce(cls, type_func, value):
        if value is None or isinstance(value, type_func):
            return value
        if issubclass(type_func, pydantic.BaseModel):
            if isinstance(value, Mapping):
                return type_func(**value)
        return type_func(value)

    def resolve(self,
                param: inspect.Parameter,
                func: Callable) -> Optional[Tuple[str, Callable]]:
        annotation = param.annotation
        key = '%s:%s' % (annotation.__name__.lower(), param.name)

        if annotation is inspect.Parameter.empty:
            key = 'empty:' + param.name
            return key, self.empty
        elif issubclass(annotation, (str, int, float, bool, _QueryData)):
            return key, self.url_or_query_argument
        elif issubclass(annotation, (dict, list, _BodyData)):
            return key, self.body_argument
        elif issubclass(annotation, _FormData):
            return key, self.form_argument
        return None

    def url_argument(self,
                     name: ParamName,
                     kwargs: KeywordArgs,
                     coerce: ParamAnnotation) -> Any:
        return self.coerce(coerce, kwargs[name])

    def query_argument(self,
                       name: ParamName,
                       query_params: http.QueryParams,
                       coerce: ParamAnnotation) -> Any:
        value = query_params.get(name)
        if issubclass(coerce, _QueryData):
            value = query_params
        return self.coerce(coerce, value)

    def form_argument(self,
                      data: http.RequestData,
                      name: ParamName,
                      coerce: ParamAnnotation) -> Any:
        if not isinstance(data, dict):
            raise exceptions.ValidationError(
                detail='Request data must be an object.'
            )
        return self.coerce(coerce, data)

    def body_argument(self,
                      data: http.RequestData,
                      coerce: ParamAnnotation) -> Any:
        return self.coerce(coerce, data)


class CoreAPISchema(Schema):

    def __init__(self,
                 router: Router,
                 routes: RouteConfig,
                 settings: Settings) -> None:
        try:
            url = router.reverse_url('serve_schema')
        except exceptions.NoReverseMatch:
            url = None

        content = {}
        for route in flatten_routes(routes):
            if getattr(route.view, 'exclude_from_schema', False):
                continue
            content[route.name] = self.get_link(route)

        schema_settings = settings.get('SCHEMA', {})
        title = schema_settings.get('TITLE', '')
        description = schema_settings.get('DESCRIPTION', '')
        url = schema_settings.get('URL', url)

        super().__init__(title=title, description=description, url=url, content=content)

    @classmethod
    def get_link(cls, route: Route) -> coreapi.Link:
        path, method, view, name = route

        fields: List[coreapi.Field] = []
        path_names = set(uritemplate.URITemplate(path).variable_names)
        for param in inspect.signature(view).parameters.values():
            fields += cls.get_field(param, path_names)

        description = cls.get_docstring(view, None)
        return coreapi.Link(url=path, action=method, description=description, fields=fields)

    @classmethod
    def get_field(cls,
                  param: inspect.Parameter,
                  path_names: Set[str]) -> List[coreapi.Field]:
        field_type = param.annotation
        if field_type is inspect.Signature.empty:
            field_type = str

        if not inspect.isclass(field_type):
            return []  # TODO callable

        if param.name in path_names:
            return [coreapi.Field(
                name=param.name,
                location='path',
                required=True,
                schema=cls.get_param_schema(field_type)
            )]

        elif (issubclass(field_type, pydantic.BaseModel)
                and issubclass(field_type, HTTPParameterData)):
            # TODO recursive models
            # TODO Optional and Union types
            return [
                coreapi.Field(
                    name=name,
                    location=field_type._schema_location,
                    required=True,
                    schema=cls.get_param_schema(value)
                )
                for name, value in field_type.__annotations__.items()
            ]

        return []

    @classmethod
    def get_param_schema(cls, annotated_type: Type) -> coreschema.schemas.Schema:
        schema_kwargs = {
            'description': cls.get_docstring(annotated_type, '')
        }
        if issubclass(annotated_type, bool):
            return coreschema.Boolean(**schema_kwargs)
        elif issubclass(annotated_type, int):
            return coreschema.Integer(**schema_kwargs)
        elif issubclass(annotated_type, float):
            return coreschema.Number(**schema_kwargs)
        elif issubclass(annotated_type, enum.Enum):
            choices = cast(Type[enum.Enum], annotated_type)
            choices = [c.value for c in choices]
            return coreschema.Enum(enum=choices, **schema_kwargs)
        return coreschema.String(**schema_kwargs)

    @classmethod
    def get_docstring(cls, obj: Type, default: Optional[str]) -> str:
        if obj.__doc__:
            return textwrap.dedent(obj.__doc__).strip()
        else:
            return default


class WSGIApp(WSGIApp):

    BUILTIN_COMPONENTS = [
        Component(Schema, init=CoreAPISchema),
        Component(Templates, init=templates.Jinja2Templates),
        Component(StaticFiles, init=statics.WhiteNoiseStaticFiles),
        Component(Router, init=router.WerkzeugRouter),
        Component(CommandLineClient, init=commandline.ArgParseCommandLineClient),
        Component(Console, init=console.PrintConsole),
        Component(SessionStore, init=sessions.LocalMemorySessionStore),
    ]

    def create_http_injector(self) -> Injector:
        http_components = {
            component.cls: component.init
            for component in self.HTTP_COMPONENTS
        }

        return self.INJECTOR_CLS(
            components={**http_components, **self.components},
            initial_state=self.preloaded_state,
            required_state={
                WSGIEnviron: 'wsgi_environ',
                Handler: 'handler',
                KeywordArgs: 'kwargs',
                Exception: 'exc',
                http.ResponseHeaders: 'response_headers',
                http.ResponseData: 'response_data'
            },
            resolvers=[PydanticHTTPResolver()]
        )


class ASyncIOApp(ASyncIOApp):

    BUILTIN_COMPONENTS = [
        Component(Schema, init=CoreAPISchema),
        Component(Templates, init=templates.Jinja2Templates),
        Component(StaticFiles, init=statics.WhiteNoiseStaticFiles),
        Component(Router, init=router.WerkzeugRouter),
        Component(CommandLineClient, init=commandline.ArgParseCommandLineClient),
        Component(Console, init=console.PrintConsole),
        Component(SessionStore, init=sessions.LocalMemorySessionStore),
    ]

    def create_http_injector(self) -> Injector:
        """
        Create the dependency injector for running handlers in response to
        incoming HTTP requests.
        """
        http_components = {
            component.cls: component.init
            for component in self.HTTP_COMPONENTS
        }

        return self.INJECTOR_CLS(
            components={**http_components, **self.components},
            initial_state=self.preloaded_state,
            required_state={
                UMIMessage: 'message',
                UMIChannels: 'channels',
                KeywordArgs: 'kwargs',
                Handler: 'handler',
                Exception: 'exc',
                http.ResponseHeaders: 'response_headers',
                http.ResponseData: 'response_data'
            },
            resolvers=[PydanticHTTPResolver()]
        )


class JSONRenderer(JSONRenderer):

    @staticmethod
    def default(obj):
        if isinstance(obj, pydantic.BaseModel):
            return obj.dict()
        if isinstance(obj, datetime.datetime):
            return obj.timestamp()
        raise TypeError(
            f"Object of type '{obj.__class__.__name__}' is not JSON serializable"
        )

    def render(self, data: http.ResponseData) -> bytes:
        return json.dumps(data, default=self.default).encode('utf-8')



#!/usr/bin/env python3

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

setup(
    name='apistar-pydantic',
    version='0.0.3',
    description="pydantic support for APIStar",
    long_description=readme,
    author="Pedro Sousa Lacerda",
    author_email='pslacerda@gmail.com',
    url='https://github.com/pslacerda/apistar_pydantic',
    py_modules=['apistar_pydantic'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'apistar',
        'pydantic',
        'coreapi',
        'coreschema',
        'uritemplate'
    ],
    license="MIT license",
    keywords='apistar_pydantic',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=[
        'pytest'
    ],
)
